@@ -66,6 +66,10 @@ class BlobReader:
     def upload_output_blob(self, blob_name, content):
         out_blob = self.get_output_blob(blob_name)
         out_blob.upload_blob(content, timeout=1200, overwrite=True)
+
+    def upload_input_blob(self, blob_name, content):
+        in_blob = self.get_input_blob(blob_name)
+        in_blob.upload_blob(content, timeout=1200, overwrite=True)
     
     def get_input_blob(self, blob_name):
         return BlobClient.from_connection_string(self.connection_string, container_name=self.input_container, blob_name=blob_name)
@@ -73,20 +77,6 @@ class BlobReader:
     def get_output_blob(self, blob_name):
         return BlobClient.from_connection_string(self.connection_string, container_name=self.output_container, blob_name=blob_name)
 
-    def read_next_batch(self, blob_name, offset, segment_count):
-        segment_id = -1
-        next_batch = []
-        blobs_iterator = self.read_blob_by_segments(blob_name)
-        for segment in blobs_iterator:
-            segment_id += 1
-            # Skip all segments until required position
-            if segment_id < offset:
-                continue
-            next_batch.append(segment)
-            if len(next_batch) == segment_count:
-                break
-        return next_batch
-    
     @staticmethod
     def _preprocess(text):
         return text.replace("\xa0", " ").replace("&nbsp;", " ").strip()
@@ -97,7 +87,8 @@ class IdGenerator:
     def generate_request_id(data, guid=None):
         if not guid:
             guid = IdGenerator.generate_id()
-        request_hash = str(hashlib.md5(data.encode()).hexdigest())
+        data_to_hash = data + str(uuid.uuid4())
+        request_hash = str(hashlib.md5(data_to_hash.encode()).hexdigest())
         request_id = "%s_%s" % (guid, request_hash)
         return guid, request_hash, request_id
 
@@ -335,17 +326,17 @@ class Worker:
     def _process_blob(self, request_id: str, metadata: TaskMetadata, func):
         try:
             if metadata.next_batch:
-                result, failures, next_batch = self._process_next_batch(metadata, func)
+                result, failures = self._process_next_batch(metadata, func)
             else:
-                result, failures, next_batch = self._process_first_batch(metadata, func)
+                result, failures = self._process_first_batch(metadata, func)
         except Exception as e:
             self._handle_exception_during_precessing(e, metadata, request_id)
             return
 
         if failures:
             self.metadata_repository.record_segment_failures(request_id, self.worker_name, failures)
-        if next_batch:
-            self._move_task_to_next_stage(metadata, result, next_batch, request_id)
+        if metadata.next_batch is not None:
+            self._move_task_to_next_stage(metadata, result, request_id)
         else:
             self._finalize_task(metadata, result)
 
@@ -357,14 +348,14 @@ class Worker:
             self.metadata_repository.update_metadata(metadata)
             self.message_queue_client.enqueue(request_id)
 
-    def _move_task_to_next_stage(self, metadata: TaskMetadata, result: list, next_batch: list, request_id: str):
+    def _move_task_to_next_stage(self, metadata: TaskMetadata, result: list, request_id: str):
         output_blob = "%s/%s_%d" % (self.worker_name, metadata.blob_name, metadata.batch_count)
         serialized_result = "\n".join(json.dumps(r) for r in result)
         self.reader.upload_output_blob(output_blob, serialized_result)
 
         metadata.offset += len(result)
         metadata.batch_count += 1
-        metadata.next_batch = next_batch
+        
         self.metadata_repository.update_metadata(metadata)
         self.message_queue_client.enqueue(request_id)
 
@@ -391,40 +382,51 @@ class Worker:
         for segment in segments_iterator:
             batch.append(segment)
             if self.segment_limit is not None and len(batch) == self.segment_limit:
-                result, failures = func(batch, metadata.params)
-                batch = []
                 break
 
         if len(batch) > 0: 
-            # If segments in blob are less than segment_limit or there is no segment limit at all
             result, failures = func(batch, metadata.params)
 
         if self.segment_limit is None:
             # If there is no segment limit, then blob is processed in one batch
             return result, failures, []
             
-        next_batch = []
+        batch = []
+        batch_id = 0
         for segment in segments_iterator:
-            next_batch.append(segment)
-            if len(next_batch) == self.segment_limit:
-                break
+            batch.append(segment)
+            if len(batch) == self.segment_limit:
+                batch_blob_name = self._get_batch_blob_name(metadata, batch_id)
+                self.reader.upload_input_blob(batch_blob_name, json.dumps(batch, ensure_ascii=False))
+                batch_id += 1
+                batch = []
 
-        return result, failures, next_batch
+        if len(batch) > 0:
+            batch_blob_name = self._get_batch_blob_name(metadata, batch_id)
+            self.reader.upload_input_blob(batch_blob_name, json.dumps(batch, ensure_ascii=False))
+            batch_id += 1
+
+        if batch_id > 0:
+            metadata.next_batch = self._get_batch_blob_name(metadata, 0)
+            metadata.total_batches = batch_id
+
+        return result, failures
 
     def _process_next_batch(self, metadata: TaskMetadata, func):
-        # Start reading new batch asynchrounously
-        batch_read_task = self.pool.apply_async(self.reader.read_next_batch, args=(
-            metadata.blob_name, 
-            metadata.offset + len(metadata.next_batch), 
-            self.segment_limit 
-        )) 
-
-        # Process current batch
-        result, failures = func(metadata.next_batch, metadata.params)
+        segments_to_process = self.reader.read_input_blob(metadata.next_batch)
+        result, failures = func(json.loads(segments_to_process), metadata.params)
         
         # Set correct segment indeces, based on offset
         for i in range(len(failures)):
             s, e, reason = failures[i]
             failures[i] = (s + metadata.offset, e + metadata.offset, reason)
-        next_batch = batch_read_task.get()
-        return result, failures, next_batch
+
+        if metadata.batch_count != metadata.total_batches:
+            metadata.next_batch = self._get_batch_blob_name(metadata, metadata.batch_count)
+        else:
+            metadata.next_batch = None
+        return result, failures
+
+    def _get_batch_blob_name(self, metadata: TaskMetadata, batch_id: int) -> str:
+        blob_folder = metadata.blob_name.split('/')[0]
+        return f"{blob_folder}/{metadata.rk}_{self.segment_limit}/{batch_id}"
